#include <SPI.h>
#include <mcp_can.h>      // https://github.com/Seeed-Studio/CAN_BUS_Shield
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

const int pcbIDsofware = 2;   // C'est le programme pour la carte CAN ayant l'ID = 2

const int pinIDhigh = 3;        // Inter IDhigh connecté à D3
const int pinIDlow = 4;         // Inter IDlow connecté à D4
const int pinLed1 = 5;          // LED1 connectée à D5
const int pinLed2 = 6;          // LED2 connectée à D6
const int pinLed3 = 7;          // LED3 connectée à D7
const int pinLed4 = 8;          // LED4 connectée à D8
const int pinSound = 9;         // Haut-parleur connectée à D9
const int analogPinPot1 = A0;   // Potentiomètre 1 connecté à A0
const int analogPinPot2 = A1;   // Potentiomètre 2 connecté à A1
const int pinBP1 = A2;          // BP1 connecté sur A2
const int pinBP2 = A3;          // BP2 connecté sur A3
const int pinBP3 = A6;          // BP3 connecté sur A6
const int pinBP4 = A7;          // BP4 connecté sur A7

const byte standardFrame = 0;
const byte extendedFrame = 1;
const byte msgLength = 8;

const byte cmdeAllumerLed = 0;
const byte cmdeEteindreLed = 1;
const byte cmdeSetAffichageA = 2;
const byte cmdeSetAffichageB = 3;
const byte cmdeEffacerEcran = 4;

const byte paramBP = 0;
const byte paramPotentiometre = 1;

const byte numPotA = 1;
const byte numPotB = 2;

const int soundSpeed = 35;

int canID;
int valPot1;
int valPot2;
int etatBPs[4];
int valBPs[4];

unsigned char msg[8] = {0, 0, 0, 0, 0, 0, 0, 0};

const int SPI_CS_PIN = 10;
MCP_CAN CAN(SPI_CS_PIN);            // 10=D10 = broche CS sur SPI pour le MCP2515
LiquidCrystal_I2C lcd(0x27,20,4);   // 0x3F pour version PCF8574-A et 0x27 pour version PCF8574-T

// Allume puis éteint les LEDS au démarrage, pour vérifier leur bon état de fonctionnement
void checkUpLeds() {
  digitalWrite(pinLed1, HIGH);
  digitalWrite(pinLed2, HIGH);
  digitalWrite(pinLed3, HIGH);
  digitalWrite(pinLed4, HIGH);
  delay(300);
  digitalWrite(pinLed1, LOW);
  digitalWrite(pinLed2, LOW);
  digitalWrite(pinLed3, LOW);
  digitalWrite(pinLed4, LOW);
}

// Envoi 3 bips courts
void bipSonoreOK() {
    for (int i = 0; i < 3; i++) {
      tone(pinSound, 1000, soundSpeed*1.5);
      delay(soundSpeed*3);
    }
}

// Envoi 1 bip long
void bipSonoreERREUR() {
    tone(pinSound, 1000, soundSpeed*20);
}

void setup()
{
    // Initialisation du port série
    Serial.begin(9600);

    // Configuration des pins
    pinMode(pinIDlow, INPUT);
    pinMode(pinIDhigh, INPUT);
    pinMode(pinLed1, OUTPUT);
    pinMode(pinLed2, OUTPUT);
    pinMode(pinLed3, OUTPUT);
    pinMode(pinLed4, OUTPUT);

    // Initialisation de la valeur des BPs au démarrage
    etatBPs[0] = 0;
    etatBPs[1] = 0;
    etatBPs[2] = 0;
    etatBPs[3] = 0;

    // Vérification bon fonctionnement LEDS
    checkUpLeds();

    // Initialisation de l'écran LCD
    lcd.init();
    lcd.backlight();
  
    // Détermination de l'ID
    canID = !digitalRead(pinIDlow) + 2*!digitalRead(pinIDhigh);
    if(canID != pcbIDsofware) {
      Serial.println("Arrêt du programme, car l'ID du programme ne correspond pas à la carte");
      Serial.print("ID software = ");
      Serial.println(canID);
      Serial.print("ID de la carte sur le BUS CAN = ");
      Serial.println(pcbIDsofware);
      bipSonoreERREUR();
      lcd.setCursor(0, 0);   lcd.print("ERREUR PARAMETRAGE");       // PosX (ligne) et PosY (colonne), en démarrant de 0
      lcd.setCursor(0, 1);   lcd.print("ID carte/logiciel");
      lcd.setCursor(0, 2);   lcd.print("differents [KO]");
      lcd.setCursor(0, 3);   lcd.print("IDlog=");
      lcd.setCursor(7, 3);   lcd.print(pcbIDsofware);
      lcd.setCursor(10, 3);  lcd.print("IDcarte=");
      lcd.setCursor(19, 3);  lcd.print(canID);
      while(1) {    
      }
    }
    delay(100);
    
    // Préparation du BUS CAN
    while (CAN_OK != CAN.begin(CAN_500KBPS))
    {
        delay(100);
    }
    Serial.println("[OK] BUS CAN initialisé !");
    Serial.println("");
    lcd.setCursor(0, 0);   lcd.print("BUS CAN [OK]");       // PosX (ligne) et PosY (colonne), en démarrant de 0
    lcd.setCursor(0, 1);   lcd.print("Carte ID = ");
    lcd.setCursor(11, 1);   lcd.print(canID);

    // init_Mask(unsigned char num, unsigned char ext, unsigned char ulData);
    // init_Filt(unsigned char num, unsigned char ext, unsigned char ulData);
              // num represents which register to use. You can fill 0 or 1 for mask and 0 to 5 for filter.
              // ext represents the status of the frame. 0 means it's a mask or filter for a standard frame. 1 means it's for a extended frame.
              // ulData represents the content of the mask of filter.

    // Bips sonores pour dire que tout est OK
    bipSonoreOK();
}

// Envoi message BPx appuyé/relâché à la carte 1
// --> dire d'allumer/d'éteindre la LED x/4
void envoiMsgSuiteBP1appuye(int BPnum, bool bAllumer)
{
  // Initialisation
  memset(msg, 0, sizeof msg);   // Vidange du msg précédent
  
  // Construction du message
  Serial.print("[BP ");
  Serial.print(BPnum);  // N° du BP
  if(bAllumer)  
    Serial.print(" appuyé] Message pour carte #1 : allumer LED n°");
  else
    Serial.print(" relâché] Message pour carte #1 : éteindre LED n°");
  Serial.print(BPnum);  // N° de la LED correspondante
  Serial.println(""); 
  msg[0] = 1;                   // ID de la cible, ou 255 si broadcast
  if (bAllumer)
    msg[1] = cmdeAllumerLed;    // Donne l'ordre d'allumer une LED
  else
    msg[1] = cmdeEteindreLed;   // Donne l'ordre d'éteindre une LED
  msg[2] = BPnum;               // Indique le n° de la LED cible
  
  // Envoi du message
  Serial.print("Message envoyé sur le BUS CAN : ");
  Serial.print(msg[0]); Serial.print(" "); Serial.print(msg[1]); Serial.print(" "); Serial.print(msg[2]); Serial.print(" "); Serial.print(msg[3]); Serial.print(" ");
  Serial.print(msg[4]); Serial.print(" "); Serial.print(msg[5]); Serial.print(" "); Serial.print(msg[6]); Serial.print(" "); Serial.print(msg[7]); Serial.print(" ");
  Serial.println("");
  CAN.sendMsgBuf(canID, standardFrame, msgLength, msg);

  // Fin de cet envoi
}

void loop()
{   
  // Scanne tous les BP, pour envoyer leur état
  valBPs[0] = abs(analogRead(pinBP1)-1023);         // Inversion des états des BPs, car reliés à une résistance de pull-down
  valBPs[1] = abs(analogRead(pinBP2)-1023);
  valBPs[2] = abs(analogRead(pinBP3)-1023);
  valBPs[3] = abs(analogRead(pinBP4)-1023);
  for(int i=0; i<4; i++) {
    if((valBPs[i]>500) && (etatBPs[i] == 0)) {
      envoiMsgSuiteBP1appuye(i+1, true);     
      etatBPs[i] = 1;
    }
    else if ((valBPs[i]<500) && (etatBPs[i] == 1)) {
      envoiMsgSuiteBP1appuye(i+1, false);
      etatBPs[i] = 0;
    }
  }

  // Boucle de traitement des messages, s'il y en a en attente
  if(CAN_MSGAVAIL == CAN.checkReceive())
  {
      unsigned char msgLongueur = 0;
      unsigned char msgBuffer[8];
      
      // Lecture du message
      CAN.readMsgBuf(&msgLongueur, msgBuffer);

      // Récupération de l'ID de l'émetteur
      unsigned long canId = CAN.getCanId();
      Serial.print("[RECEPTION DONNEES] ID emetteur = 0x"); Serial.println(canId, HEX);

      // Affichage du message sur le port sériel
      for(int i = 0; i<msgLongueur; i++)
      {
          Serial.print(msgBuffer[i]);
          Serial.print("\t");
      }
      Serial.println();

      // Traitement du message reçu
      if((byte)msgBuffer[0] == (byte)canID) {              // On vérifie qu'on s'adresse bien à nous (ici ID carte = 2)
        if ((byte)msgBuffer[1] == (byte)cmdeSetAffichageA) {      // Demande d'écriture blocA de l'afficheur (3ème ligne, ici)
          if((byte)msgBuffer[2] == (byte)paramPotentiometre) {            // Regarde si ce que l'on souhaite afficher est la valeur d'un potentiomètre
            if((byte)msgBuffer[3] == numPotA) {   // PotA
              int valPotA = ((int)msgBuffer[4])*256 + (int)msgBuffer[5];
              lcd.setCursor(0, 2);   lcd.print("PotA=    ");       // PosX (ligne) et PosY (colonne), en démarrant de 0
              lcd.setCursor(5, 2);   lcd.print(valPotA);
            }
            if((byte)msgBuffer[3] == numPotB) {   // PotB
              int valPotB = ((int)msgBuffer[4])*256 + (int)msgBuffer[5];
              lcd.setCursor(11, 2);   lcd.print("PotB=    ");       // PosX (ligne) et PosY (colonne), en démarrant de 0
              lcd.setCursor(16, 2);   lcd.print(valPotB);
            }           
          }
        }
     }
  }

  // Loop
}
